from datetime import tzinfo
from pathlib import Path
from typing import Any, Optional

from allotropy.allotrope.allotrope import serialize_allotrope
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parser_factory import get_parser, VendorType
from allotropy.types import IOType


def allotrope_from_io(
    contents: IOType,
    filename: str,
    vendor_type: VendorType,
    default_timezone: Optional[tzinfo] = None,
) -> dict[str, Any]:
    return serialize_allotrope(
        allotrope_model_from_io(contents, filename, vendor_type, default_timezone)
    )


def allotrope_model_from_io(
    contents: IOType,
    filename: str,
    vendor_type: VendorType,
    default_timezone: Optional[tzinfo] = None,
) -> Any:
    named_file_contents = NamedFileContents(contents, filename)
    parser = get_parser(vendor_type, default_timezone=default_timezone)
    return parser.to_allotrope(named_file_contents)


def allotrope_from_file(
    filepath: str,
    vendor_type: VendorType,
    default_timezone: Optional[tzinfo] = None,
) -> dict[str, Any]:
    try:
        with open(filepath, "rb") as f:
            return allotrope_from_io(
                f, Path(filepath).name, vendor_type, default_timezone=default_timezone
            )
    except FileNotFoundError as e:
        msg = f"File not found: {filepath}."
        raise AllotropeConversionError(msg) from e


def allotrope_model_from_file(
    filepath: str,
    vendor_type: VendorType,
    default_timezone: Optional[tzinfo] = None,
) -> Any:
    with open(filepath, "rb") as f:
        return allotrope_model_from_io(
            f, Path(filepath).name, vendor_type, default_timezone=default_timezone
        )



import itertools
from typing import Any, Union
import uuid

from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    ContainerType as FluorescenceContainerType,
    MeasurementAggregateDocument as FluorescenceMeasurementAggregateDocument,
    Model as FluorescenceModel,
)
from allotropy.allotrope.models.luminescence_benchling_2023_09_luminescence import (
    ContainerType as LuminescenceContainerType,
    MeasurementAggregateDocument as LuminescenceMeasurementAggregateDocument,
    Model as LuminescenceModel,
)
from allotropy.allotrope.models.shared.definitions.custom import TQuantityValueNumber
from allotropy.allotrope.models.ultraviolet_absorbance_benchling_2023_09_ultraviolet_absorbance import (
    ContainerType as AbsorbanceContainerType,
    MeasurementAggregateDocument as AbsorbanceMeasurementAggregateDocument,
    Model as AbsorbanceModel,
)
from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import Data
from allotropy.parsers.agilent_gen5.constants import ReadMode
from allotropy.parsers.agilent_gen5.plate_data import PlateData
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader
from allotropy.parsers.vendor_parser import VendorParser


class AgilentGen5Parser(VendorParser):
    def _create_model(
        self, first_plate: PlateData, measurement_docs: list[Any]
    ) -> Union[AbsorbanceModel, FluorescenceModel, LuminescenceModel]:
        if first_plate.plate_type.read_mode == ReadMode.ABSORBANCE:
            return AbsorbanceModel(
                measurement_aggregate_document=AbsorbanceMeasurementAggregateDocument(
                    measurement_identifier=str(uuid.uuid4()),
                    measurement_time=self._get_date_time(
                        first_plate.plate_number.datetime
                    ),
                    analytical_method_identifier=first_plate.file_paths.protocol_file_path,
                    experimental_data_identifier=first_plate.file_paths.experiment_file_path,
                    container_type=AbsorbanceContainerType.well_plate,
                    plate_well_count=TQuantityValueNumber(
                        len(first_plate.results.wells)
                    ),
                    # TODO read_type=self.read_type.value?,
                    measurement_document=measurement_docs,
                )
            )
        elif first_plate.plate_type.read_mode == ReadMode.FLUORESCENCE:
            return FluorescenceModel(
                measurement_aggregate_document=FluorescenceMeasurementAggregateDocument(
                    measurement_identifier=str(uuid.uuid4()),
                    measurement_time=self._get_date_time(
                        first_plate.plate_number.datetime
                    ),
                    analytical_method_identifier=first_plate.file_paths.protocol_file_path,
                    experimental_data_identifier=first_plate.file_paths.experiment_file_path,
                    container_type=FluorescenceContainerType.well_plate,
                    plate_well_count=TQuantityValueNumber(
                        len(first_plate.results.wells)
                    ),
                    # TODO read_type=self.read_type.value?,
                    measurement_document=measurement_docs,
                )
            )
        elif first_plate.plate_type.read_mode == ReadMode.LUMINESCENCE:
            return LuminescenceModel(
                measurement_aggregate_document=LuminescenceMeasurementAggregateDocument(
                    measurement_identifier=str(uuid.uuid4()),
                    measurement_time=self._get_date_time(
                        first_plate.plate_number.datetime
                    ),
                    analytical_method_identifier=first_plate.file_paths.protocol_file_path,
                    experimental_data_identifier=first_plate.file_paths.experiment_file_path,
                    container_type=LuminescenceContainerType.well_plate,
                    plate_well_count=TQuantityValueNumber(
                        len(first_plate.results.wells)
                    ),
                    # TODO read_type=self.read_type.value?,
                    measurement_document=measurement_docs,
                )
            )

        msg = msg_for_error_on_unrecognized_value(
            "read mode", first_plate.plate_type.read_mode, ReadMode._member_names_
        )
        raise AllotropeConversionError(msg)

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Any:
        contents = named_file_contents.contents
        section_lines_reader = SectionLinesReader(contents, encoding=None)
        data = Data.create(section_lines_reader)

        first_plate = data.plates[0]
        # TODO we just use the metadata for the first plate, but in theory they could all have
        # different metadata in particular, timestamp is known to be, but not sure on procedures/
        # plate dimensions/etc., need to follow up
        measurement_docs = list(
            itertools.chain.from_iterable(
                [plate.results.measurement_docs for plate in data.plates]
            )
        )

        return self._create_model(first_plate, measurement_docs)

        # TODO stats docs
        # statistics_docs = [plate.statistics_doc for plate in plates]
        # all_statistics_docs = list(itertools.chain.from_iterable(statistics_docs))
        # if all_statistics_docs:
        #     allotrope_dict["measurement aggregate document"]["statistics aggregate document"] = {
        #         "statistics document": all_statistics_docs,
        #     }



from __future__ import annotations

from dataclasses import dataclass

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.agilent_gen5.plate_data import PlateData
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader


@dataclass(frozen=True)
class Data:
    plates: list[PlateData]

    @staticmethod
    def create(section_lines_reader: SectionLinesReader) -> Data:
        plates: list[PlateData] = [
            PlateData.create(lines_reader)
            for lines_reader in section_lines_reader.iter_sections("^Software Version")
        ]

        if not plates:
            msg = "No plate data found in file."
            raise AllotropeConversionError(msg)

        return Data(plates)



import re

WAVELENGTH_COLUMNS_RE = re.compile(r"^A\d{3}$")
NO_WAVELENGTH_COLUMN_ERROR_MSG = (
    "The file is required to include an absorbance measurement column (e.g. A280)"
)
INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG = (
    msg
) = "The wavelength_column should follow the pattern A### (ex. A260)."
NO_DATE_OR_TIME_ERROR_MSG = (
    "The file is required to include both 'Date' and 'Time' columns."
)
NO_MEASUREMENT_IN_PLATE_ERROR_MSG = (
    "The plate data does not contain absorbance measurement for {}."
)



# mypy: disallow_any_generics = False

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime as datetime_lib
from io import StringIO
from typing import Optional, Union

import numpy as np
import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TDatacube,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
)
from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.parsers.agilent_gen5.absorbance_data_point import AbsorbanceDataPoint
from allotropy.parsers.agilent_gen5.constants import (
    ReadMode,
    ReadType,
    READTYPE_TO_DIMENSIONS,
)
from allotropy.parsers.agilent_gen5.data_point import DataPoint
from allotropy.parsers.agilent_gen5.fluorescence_data_point import FluorescenceDataPoint
from allotropy.parsers.agilent_gen5.luminescence_data_point import LuminescenceDataPoint
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.values import assert_not_none

METADATA_PREFIXES = frozenset(
    {
        "Plate Number",
        "Date",
        "Time",
        "Reader Type:",
        "Reader Serial Number:",
        "Reading Type",
    }
)
GEN5_DATETIME_FORMAT = "%m/%d/%Y %I:%M:%S %p"


def try_float(value: str) -> Union[str, float]:
    try:
        return float(value)
    except ValueError:
        return value


def read_data_section(lines_reader: LinesReader) -> str:
    data_section = "\n".join(
        [
            lines_reader.pop() or "",
            *lines_reader.pop_until_empty(),
        ]
    )
    lines_reader.drop_empty()
    return data_section


def hhmmss_to_sec(hhmmss: str) -> int:
    # convert to seconds, as specified by Allotrope
    hours, minutes, seconds = tuple(int(num) for num in hhmmss.split(":"))
    return (3600 * hours) + (60 * minutes) + seconds


@dataclass(frozen=True)
class FilePaths:
    experiment_file_path: str
    protocol_file_path: str

    @staticmethod
    def create(lines_reader: LinesReader) -> FilePaths:
        assert_not_none(
            lines_reader.drop_until("^Experiment File Path"),
            "Experiment File Path",
        )
        file_paths = lines_reader.pop_until_empty()
        return FilePaths(
            experiment_file_path=f"{next(file_paths)}\t".split("\t")[1],
            protocol_file_path=f"{next(file_paths)}\t".split("\t")[1],
        )


@dataclass(frozen=True)
class PlateNumber:
    datetime: str
    plate_barcode: str

    @staticmethod
    def create(lines_reader: LinesReader) -> PlateNumber:
        assert_not_none(lines_reader.drop_until("^Plate Number"), "Plate Number")
        metadata_dict = PlateNumber._parse_metadata(lines_reader)

        return PlateNumber(
            datetime=datetime_lib.strptime(  # noqa: DTZ007
                f"{metadata_dict['Date']} {metadata_dict['Time']}",
                GEN5_DATETIME_FORMAT,
            ).isoformat(),
            plate_barcode=metadata_dict["Plate Number"],
        )

    @staticmethod
    def _parse_metadata(lines_reader: LinesReader) -> dict:
        metadata_dict: dict = {}
        for metadata_line in lines_reader.pop_until_empty():
            line_split = metadata_line.split("\t")
            if line_split[0] not in METADATA_PREFIXES:
                msg = msg_for_error_on_unrecognized_value(
                    "metadata key", line_split[0], METADATA_PREFIXES
                )
                raise AllotropeConversionError(msg)
            metadata_dict[line_split[0]] = line_split[1]
        # TODO put more metadata in the right spots
        return metadata_dict


@dataclass(frozen=True)
class PlateType:
    read_mode: ReadMode
    read_type: ReadType
    read_names: list[str]

    @staticmethod
    def create(lines_reader: LinesReader) -> PlateType:
        assert_not_none(lines_reader.drop_until("^Plate Type"), "Plate Type")
        data_section = read_data_section(lines_reader)

        read_mode = PlateType.get_read_mode(data_section)
        read_type = PlateType.get_read_type(data_section)
        read_names: list = []
        for procedure_chunk in PlateType._parse_procedure_chunks(data_section):
            read_names.extend(
                PlateType._parse_procedure_chunk(procedure_chunk, read_mode)
            )

        return PlateType(
            read_mode=read_mode,
            read_type=read_type,
            read_names=read_names,
        )

    @property
    def data_point_cls(self) -> type[DataPoint]:
        if self.read_mode == ReadMode.ABSORBANCE:
            return AbsorbanceDataPoint
        elif self.read_mode == ReadMode.FLUORESCENCE:
            return FluorescenceDataPoint
        elif self.read_mode == ReadMode.LUMINESCENCE:
            return LuminescenceDataPoint

        msg = msg_for_error_on_unrecognized_value(
            "read mode", self.read_mode, ReadMode._member_names_
        )
        raise AllotropeConversionError(msg)

    @staticmethod
    def get_read_mode(data_section: str) -> ReadMode:
        if ReadMode.ABSORBANCE.value in data_section:
            return ReadMode.ABSORBANCE
        elif ReadMode.FLUORESCENCE.value in data_section:
            return ReadMode.FLUORESCENCE
        elif ReadMode.LUMINESCENCE.value in data_section:
            return ReadMode.LUMINESCENCE

        msg = f"Read mode not found; expected to find one of {sorted(ReadMode._member_names_)}."
        raise AllotropeConversionError(msg)

    @staticmethod
    def get_read_type(procedure_details: str) -> ReadType:
        # TODO parse the rest of the procedure details
        if ReadType.KINETIC.value in procedure_details:
            return ReadType.KINETIC
        elif ReadType.AREASCAN.value in procedure_details:
            return ReadType.AREASCAN
        elif ReadType.SPECTRAL.value in procedure_details:
            return ReadType.SPECTRAL

        # check for this last, because other modes still contain the word "Endpoint"
        return ReadType.ENDPOINT

    @staticmethod
    def _parse_procedure_chunks(procedure_details: str) -> list[list[str]]:
        procedure_chunks = []
        current_chunk: list[str] = []
        procedure_lines = procedure_details.splitlines()
        for procedure_line in procedure_lines:
            if procedure_line[0] != "\t":
                procedure_chunks.append(current_chunk)
                current_chunk = []
            current_chunk.append(procedure_line)
        del procedure_chunks[0]
        procedure_chunks.append(current_chunk)

        return procedure_chunks

    @staticmethod
    def _parse_procedure_chunk(
        procedure_chunk: list[str],
        read_mode: ReadMode,
    ) -> list[str]:
        # if no user-defined name is specified for protocols,
        # e.g. it just says "Absorbance Endpoint",
        # Gen5 defaults to using the wavelength as the name
        read_names = []
        use_wavelength_names = False
        read_line_length = 2
        wavelength_line_length = 2
        for line in procedure_chunk:
            split_line = line.strip().split("\t")
            if split_line[0] == "Read":
                if len(split_line) != read_line_length:
                    msg = f"Expected the Read data line {split_line} to contain exactly {read_line_length} values."
                    raise AllotropeConversionError(msg)
                if split_line[-1] == f"{read_mode.title()} Endpoint":
                    use_wavelength_names = True
                else:
                    read_names.append(split_line[-1])
            elif split_line[0].startswith("Wavelengths"):
                if use_wavelength_names:
                    split_line_colon = split_line[0].split(":  ")
                    if len(split_line_colon) != wavelength_line_length:
                        msg = f"Expected the Wavelengths data line {split_line} to contain exactly {wavelength_line_length} values."
                        raise AllotropeConversionError(msg)
                    read_names.extend(split_line_colon[-1].split(", "))
        return read_names


@dataclass(frozen=True)
class LayoutData:
    layout: dict
    concentrations: dict

    @staticmethod
    def create_default() -> LayoutData:
        return LayoutData(
            layout={},
            concentrations={},
        )

    @staticmethod
    def create(layout_str: str) -> LayoutData:
        layout = {}
        concentrations = {}

        layout_lines = layout_str.splitlines()
        # first line is "Layout", second line is column numbers
        current_row = "A"
        for i in range(2, len(layout_lines)):
            split_line = layout_lines[i].split("\t")
            if split_line[0]:
                current_row = split_line[0]
            label = split_line[-1]
            for j in range(1, len(split_line) - 1):
                well_loc = f"{current_row}{j}"
                if label == "Well ID":
                    layout[well_loc] = split_line[j]
                elif label == "Conc/Dil" and split_line[j]:
                    concentrations[well_loc] = float(split_line[j])

        return LayoutData(
            layout=layout,
            concentrations=concentrations,
        )


@dataclass(frozen=True)
class ActualTemperature:
    value: Optional[float] = None

    @staticmethod
    def create_default() -> ActualTemperature:
        return ActualTemperature()

    @staticmethod
    def create(actual_temperature: str) -> ActualTemperature:
        if len(actual_temperature.split("\n")) != 1:
            msg = f"Expected the Temperature section '{actual_temperature}' to contain exactly 1 line."
            raise AllotropeConversionError(msg)

        return ActualTemperature(
            value=float(actual_temperature.strip().split("\t")[-1]),
        )


@dataclass(frozen=True)
class Results:
    measurements: defaultdict[str, list]
    processed_datas: defaultdict[str, list]
    wells: list
    measurement_docs: list

    @staticmethod
    def create() -> Results:
        return Results(
            measurements=defaultdict(list),
            processed_datas=defaultdict(list),
            wells=[],
            measurement_docs=[],
        )

    def parse_results(
        self,
        results: str,
        plate_type: PlateType,
        plate_number: PlateNumber,
        layout_data: LayoutData,
        actual_temperature: ActualTemperature,
    ) -> None:
        result_lines = results.splitlines()
        if result_lines[0].strip() != "Results":
            msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
            raise AllotropeConversionError(msg)
        # result_lines[1] contains column numbers

        current_row = "A"
        for row_num in range(2, len(result_lines)):
            values = result_lines[row_num].split("\t")
            if values[0]:
                current_row = values[0]
            label = values[-1]  # last column gives information about the type of read
            for col_num in range(1, len(values) - 1):
                well_pos = f"{current_row}{col_num}"
                if well_pos not in self.wells:
                    self.wells.append(well_pos)
                well_value: Union[str, float] = try_float(values[col_num])
                if Results._is_processed_data_label(
                    label,
                    plate_type.read_mode,
                    plate_type.read_names,
                ):
                    self.processed_datas[well_pos].append([label, well_value])
                else:
                    label_only = label.split(":")[-1]
                    self.measurements[well_pos].append([label_only, well_value])

        for well_pos in self.wells:
            self.measurement_docs.append(
                plate_type.data_point_cls(
                    plate_type.read_type,
                    self.measurements[well_pos],
                    well_pos,
                    plate_number.plate_barcode,
                    layout_data.layout.get(well_pos),
                    layout_data.concentrations.get(well_pos),
                    self.processed_datas[well_pos],
                    actual_temperature.value,
                ).to_measurement_doc()
            )

    @staticmethod
    def _is_processed_data_label(
        label: str,
        read_mode: ReadMode,
        read_names: list,
    ) -> bool:
        if read_mode == ReadMode.LUMINESCENCE:
            return not any(
                (label.startswith(read_name) and label.split(":")[-1] == "Lum")
                for read_name in read_names
            )
        else:
            return not any(
                (label.startswith(read_name) and label.split(":")[-1][0].isdigit())
                for read_name in read_names
            )


@dataclass(frozen=True)
class CurveName:
    statistics_doc: list

    @staticmethod
    def create_default() -> CurveName:
        return CurveName(statistics_doc=[])

    @staticmethod
    def create(
        stdcurve: str,
        plate_number: PlateNumber,
        results: Results,
    ) -> CurveName:
        lines = stdcurve.splitlines()
        num_lines = 2
        if len(lines) != num_lines:
            msg = f"Expected the std curve data '{lines}' to contain exactly {num_lines} lines."
            raise AllotropeConversionError(msg)
        keys = lines[0].split("\t")
        values = lines[1].split("\t")
        return CurveName(
            statistics_doc=[
                {
                    "statistical feature": key,
                    "feature": try_float(value),
                    "group": f"{plate_number.plate_barcode} {results.wells[0]}-{results.wells[-1]}",
                }
                for key, value in zip(keys, values)
            ]
        )


@dataclass(frozen=True)
class KineticData:
    temperatures: list
    kinetic_times: list[int]

    @staticmethod
    def create_default() -> KineticData:
        return KineticData(
            temperatures=[],
            kinetic_times=[],
        )

    @staticmethod
    def create(
        lines_reader: LinesReader,
        results: Results,
    ) -> KineticData:
        kinetic_data = read_data_section(lines_reader)

        kinetic_data_io = StringIO(kinetic_data)
        df = pd.read_table(kinetic_data_io)
        df_columns = kinetic_data.split("\n")[0].split("\t")
        df = df[
            df["A1"].notna()
        ]  # drop incomplete rows, particularly rows only with "0:00:00"

        kinetic_times = [hhmmss_to_sec(hhmmss) for hhmmss in df["Time"]]
        temperatures = df[df_columns[1]].replace(np.nan, None).tolist()
        has_temperatures = any(temp is not None for temp in temperatures)
        for well_pos in df_columns[
            2:
        ]:  # first column is Time, second column is T∞ READ_NAME with no values
            results.wells.append(well_pos)
            values = df[well_pos].tolist()
            if has_temperatures:
                results.measurements[well_pos].extend(
                    list(zip(kinetic_times, values, temperatures))
                )
            else:
                results.measurements[well_pos].extend(list(zip(kinetic_times, values)))

        return KineticData(
            temperatures=temperatures,
            kinetic_times=kinetic_times,
        )

    def parse_blank_kinetic_data(
        self,
        lines_reader: LinesReader,
        blank_kinetic_data_label: str,
        results: Results,
        plate_type: PlateType,
    ) -> None:
        blank_kinetic_data = read_data_section(lines_reader)

        blank_kinetic_data_io = StringIO(blank_kinetic_data)
        df = pd.read_table(blank_kinetic_data_io)
        df.dropna(axis=0)  # drop incomplete rows
        df_columns = blank_kinetic_data.split("\n")[0].split("\t")

        for well_pos in df_columns[1:]:  # first column is Time
            measures = df[well_pos].tolist()
            results.processed_datas[well_pos].append(
                [
                    blank_kinetic_data_label,
                    self._blank_data_cube(
                        measures,
                        plate_type,
                    ),
                ]
            )

    def _blank_data_cube(
        self,
        measures: list[float],
        plate_type: PlateType,
    ) -> TDatacube:
        structure_dimensions = READTYPE_TO_DIMENSIONS[plate_type.read_type]
        structure_measures = [
            (
                "double",
                plate_type.read_mode.lower(),
                plate_type.data_point_cls.unit,
            )
        ]
        return TDatacube(
            label=f"{plate_type.read_type.value.lower()} data",
            cube_structure=TDatacubeStructure(
                [
                    TDatacubeComponent(FieldComponentDatatype(data_type), concept, unit)
                    for data_type, concept, unit in structure_dimensions
                ],
                [
                    TDatacubeComponent(FieldComponentDatatype(data_type), concept, unit)
                    for data_type, concept, unit in structure_measures
                ],
            ),
            data=TDatacubeData([self.kinetic_times], [measures]),  # type: ignore[list-item]
        )


@dataclass(frozen=True)
class PlateData:
    file_paths: FilePaths
    plate_number: PlateNumber
    plate_type: PlateType
    results: Results
    curve_name: CurveName
    kinetic_data: KineticData

    @staticmethod
    def create(lines_reader: LinesReader) -> PlateData:
        file_paths = FilePaths.create(lines_reader)
        plate_number = PlateNumber.create(lines_reader)
        plate_type = PlateType.create(lines_reader)
        layout_data = LayoutData.create_default()
        actual_temperature = ActualTemperature.create_default()
        results = Results.create()
        curve_name = CurveName.create_default()
        kinetic_data = KineticData.create_default()

        while lines_reader.current_line_exists():
            data_section = read_data_section(lines_reader)
            if data_section.startswith("Layout"):
                layout_data = LayoutData.create(data_section)
            elif data_section.startswith("Actual Temperature"):
                actual_temperature = ActualTemperature.create(data_section)
            elif data_section.startswith("Results"):
                results.parse_results(
                    data_section,
                    plate_type,
                    plate_number,
                    layout_data,
                    actual_temperature,
                )
            elif data_section.startswith("Curve Name"):
                curve_name = CurveName.create(
                    data_section,
                    plate_number,
                    results,
                )
            elif len(data_section.split("\n")) == 1 and any(
                read_name in data_section for read_name in plate_type.read_names
            ):
                if data_section.startswith("Blank"):
                    kinetic_data.parse_blank_kinetic_data(
                        lines_reader,
                        data_section.strip(),
                        results,
                        plate_type,
                    )
                else:
                    kinetic_data = KineticData.create(
                        lines_reader,
                        results,
                    )

        return PlateData(
            file_paths=file_paths,
            plate_number=plate_number,
            plate_type=plate_type,
            results=results,
            curve_name=curve_name,
            kinetic_data=kinetic_data,
        )



from typing import NamedTuple

from allotropy.types import IOType


class NamedFileContents(NamedTuple):
    """
    A file's contents and the name of the file whence they came.

    In order to support use cases where the file contents of a vendor file are not stored locally,
    we accept an IO object for file contents. However, some vendor file parsers extract information
    from the original file name produced by the instrument. Thus, we use a tuple of file contents
    and original file name to support as many cases as possible.
    """

    contents: IOType
    original_file_name: str



from allotropy.allotrope.models.plate_reader_benchling_2023_09_plate_reader import (
    ContainerType,
    DataSystemDocument,
    DeviceSystemDocument,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    SampleDocument,
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TRelativeFluorescenceUnit,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_structure import (
    BackgroundInfo,
    BackgroundInfoList,
    BasicAssayInfo,
    CalculatedResultList,
    Data,
    Filter,
    Instrument,
    Labels,
    Plate,
    PlateList,
    PlateMap,
    Result,
    ResultList,
    ResultPlateInfo,
    Software,
)


def get_data() -> Data:
    return Data(
        plate_list=PlateList(
            plates=[
                Plate(
                    plate_info=ResultPlateInfo(
                        number="1",
                        barcode="Plate 1",
                        emission_filter_id="1st",
                        measinfo="1st Ex=Top Em=Top Wdw=1 (14)",
                        measurement_time="10/13/2022 3:08:06 PM",
                        measured_height=11.9,
                        chamber_temperature_at_start=23.17,
                        label="AC HTRF Laser [Eu]",
                    ),
                    background_info_list=BackgroundInfoList(
                        background_info=[
                            BackgroundInfo(
                                plate_num="1",
                                label="AC HTRF Laser [Eu]",
                                measinfo="De=1st Ex=Top Em=Top Wdw=1 (14)",
                            ),
                        ],
                    ),
                    calculated_result_list=CalculatedResultList([]),
                    result_list=ResultList(
                        [
                            Result(
                                uuid="80d11e7d-734c-4506-8087-335769da996c",
                                col="A",
                                row="01",
                                value=31441,
                            )
                        ]
                    ),
                ),
                Plate(
                    plate_info=ResultPlateInfo(
                        number="1",
                        barcode="Plate 1",
                        emission_filter_id="2nd",
                        measinfo="De=2nd Ex=Top Em=Top Wdw=1 (142)",
                        measurement_time="10/13/2022 3:08:06 PM",
                        measured_height=11.9,
                        chamber_temperature_at_start=23.17,
                        label="AC HTRF Laser [Eu]",
                    ),
                    background_info_list=BackgroundInfoList(
                        background_info=[
                            BackgroundInfo(
                                plate_num="1",
                                label="AC HTRF Laser [Eu]",
                                measinfo="De=2nd Ex=Top Em=Top Wdw=1 (142)",
                            ),
                        ],
                    ),
                    calculated_result_list=CalculatedResultList([]),
                    result_list=ResultList(
                        [
                            Result(
                                uuid="f2d4dd7c-0b02-4bd6-a6c5-8acd944e8d56",
                                col="A",
                                row="01",
                                value=80368,
                            )
                        ]
                    ),
                ),
            ],
        ),
        basic_assay_info=BasicAssayInfo("100302", "3134"),
        number_of_wells=96.0,
        plate_maps={
            "1": PlateMap(
                plate_n="1",
                group_n="1",
                sample_role_type_mapping={
                    "01": {"A": SampleRoleType.unknown_sample_role}
                },
            ),
        },
        labels=Labels(
            label="AC HTRF Laser [Eu]",
            excitation_filter=Filter("UV2 (TRF) 320", 320, 75),
            emission_filters={
                "1st": Filter("APC 665", 665, 75),
                "2nd": Filter("Cy5 620", 620, 10),
            },
            scan_position_setting=ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
            number_of_flashes=50,
            detector_gain_setting="2",
        ),
        instrument=Instrument(
            serial_number="1050209",
            nickname="EnVision",
        ),
        software=Software(
            software_name="EnVision Workstation",
            software_version="1.0",
        ),
    )


def get_model() -> Model:
    return Model(
        field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
        plate_reader_aggregate_document=PlateReaderAggregateDocument(
            device_system_document=DeviceSystemDocument(
                model_number="EnVision",
                equipment_serial_number="1050209",
                device_identifier="EnVision",
            ),
            data_system_document=DataSystemDocument(
                file_name="file.txt",
                software_name="EnVision Workstation",
                software_version="1.0",
                ASM_converter_name=ASM_CONVERTER_NAME,
                ASM_converter_version=ASM_CONVERTER_VERSION,
            ),
            plate_reader_document=[
                PlateReaderDocumentItem(
                    measurement_aggregate_document=MeasurementAggregateDocument(
                        measurement_time="2022-10-13T15:08:06+00:00",
                        analytical_method_identifier="100302",
                        experimental_data_identifier="3134",
                        container_type=ContainerType.well_plate,
                        plate_well_count=TQuantityValueNumber(value=96.0),
                        measurement_document=[
                            FluorescencePointDetectionMeasurementDocumentItems(
                                measurement_identifier="",
                                device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                                    device_control_document=[
                                        FluorescencePointDetectionDeviceControlDocumentItem(
                                            device_type="fluorescence detector",
                                            shaking_configuration_description=None,
                                            detector_distance_setting__plate_reader_=TQuantityValueMillimeter(
                                                value=11.9,
                                                has_statistic_datum_role=None,
                                                unit="mm",
                                                field_type=None,
                                            ),
                                            integration_time=None,
                                            number_of_averages=TQuantityValueNumber(
                                                value=50.0,
                                                has_statistic_datum_role=None,
                                                unit="#",
                                                field_type=None,
                                            ),
                                            detector_gain_setting="2",
                                            scan_position_setting__plate_reader_=ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
                                            detector_carriage_speed_setting=None,
                                            detector_wavelength_setting=TQuantityValueNanometer(
                                                value=665.0,
                                                has_statistic_datum_role=None,
                                                unit="nm",
                                                field_type=None,
                                            ),
                                            detector_bandwidth_setting=TQuantityValueNanometer(
                                                value=75.0,
                                                has_statistic_datum_role=None,
                                                unit="nm",
                                                field_type=None,
                                            ),
                                            excitation_bandwidth_setting=TQuantityValueNanometer(
                                                value=75.0,
                                                has_statistic_datum_role=None,
                                                unit="nm",
                                                field_type=None,
                                            ),
                                            excitation_wavelength_setting=TQuantityValueNanometer(
                                                value=320.0,
                                                has_statistic_datum_role=None,
                                                unit="nm",
                                                field_type=None,
                                            ),
                                            wavelength_filter_cutoff_setting=None,
                                            field_index=None,
                                        )
                                    ]
                                ),
                                sample_document=SampleDocument(
                                    location_identifier="A01",
                                    sample_identifier="Plate 1 A01",
                                    batch_identifier=None,
                                    sample_role_type=SampleRoleType.unknown_sample_role.value,
                                    well_plate_identifier="Plate 1",
                                ),
                                compartment_temperature=TQuantityValueDegreeCelsius(
                                    value=23.17,
                                    has_statistic_datum_role=None,
                                    unit="degC",
                                    field_type=None,
                                ),
                                fluorescence=TRelativeFluorescenceUnit(31441),
                            ),
                            FluorescencePointDetectionMeasurementDocumentItems(
                                measurement_identifier="",
                                device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                                    device_control_document=[
                                        FluorescencePointDetectionDeviceControlDocumentItem(
                                            device_type="fluorescence detector",
                                            shaking_configuration_description=None,
                                            detector_distance_setting__plate_reader_=TQuantityValueMillimeter(
                                                value=11.9,
                                                has_statistic_datum_role=None,
                                                unit="mm",
                                                field_type=None,
                                            ),
                                            integration_time=None,
                                            number_of_averages=TQuantityValueNumber(
                                                value=50.0,
                                                has_statistic_datum_role=None,
                                                unit="#",
                                                field_type=None,
                                            ),
                                            detector_gain_setting="2",
                                            scan_position_setting__plate_reader_=ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
                                            detector_carriage_speed_setting=None,
                                            detector_wavelength_setting=TQuantityValueNanometer(
                                                value=620.0,
                                                has_statistic_datum_role=None,
                                                unit="nm",
                                                field_type=None,
                                            ),
                                            detector_bandwidth_setting=TQuantityValueNanometer(
                                                value=10.0,
                                                has_statistic_datum_role=None,
                                                unit="nm",
                                                field_type=None,
                                            ),
                                            excitation_bandwidth_setting=TQuantityValueNanometer(
                                                value=75.0,
                                                has_statistic_datum_role=None,
                                                unit="nm",
                                                field_type=None,
                                            ),
                                            excitation_wavelength_setting=TQuantityValueNanometer(
                                                value=320.0,
                                                has_statistic_datum_role=None,
                                                unit="nm",
                                                field_type=None,
                                            ),
                                            wavelength_filter_cutoff_setting=None,
                                            field_index=None,
                                        )
                                    ]
                                ),
                                sample_document=SampleDocument(
                                    location_identifier="A01",
                                    sample_identifier="Plate 1 A01",
                                    batch_identifier=None,
                                    sample_role_type=SampleRoleType.unknown_sample_role.value,
                                    well_plate_identifier="Plate 1",
                                ),
                                compartment_temperature=TQuantityValueDegreeCelsius(
                                    value=23.17,
                                    has_statistic_datum_role=None,
                                    unit="degC",
                                    field_type=None,
                                ),
                                fluorescence=TRelativeFluorescenceUnit(80368),
                            ),
                        ],
                    )
                ),
            ],
        ),
    )



from collections.abc import Iterator
from io import StringIO
from re import search
from typing import Optional

import chardet
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.types import IOType

EMPTY_STR_PATTERN = r"^\s*$"


def _decode(bytes_content: bytes, encoding: Optional[str]) -> str:
    if not encoding:
        encoding = chardet.detect(bytes_content)["encoding"]
        if not encoding:
            error = "Unable to detect text encoding for file. The file may be empty."
            raise AllotropeConversionError(error)
    return bytes_content.decode(encoding)


class LinesReader:
    def __init__(self, io_: IOType, encoding: Optional[str] = "UTF-8"):
        stream_contents = io_.read()
        self.raw_contents = (
            _decode(stream_contents, encoding)
            if isinstance(stream_contents, bytes)
            else stream_contents
        )
        self.contents = self.raw_contents.replace("\r\n", "\n")
        self.lines: list[str] = self.contents.split("\n")
        self.n_lines = len(self.lines)
        self.current_line = 0

    def current_line_exists(self) -> bool:
        return 0 <= self.current_line < self.n_lines

    def get(self) -> Optional[str]:
        return self.lines[self.current_line] if self.current_line_exists() else None

    def match(self, match_pat: str) -> bool:
        line = self.get()
        return False if line is None else bool(search(match_pat, line))

    def is_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> bool:
        return self.match(empty_pat)

    def pop(self) -> Optional[str]:
        line = self.get()
        if line is not None:
            self.current_line += 1
        return line

    def pop_if_match(self, match_pat: str) -> Optional[str]:
        return self.pop() if self.match(match_pat) else None

    def pop_data(self) -> Optional[str]:
        self.drop_empty()
        return self.pop()

    def drop_until(self, match_pat: str) -> Optional[str]:
        while self.current_line_exists() and not self.match(match_pat):
            self.pop()
        return self.get()

    def drop_until_inclusive(self, match_pat: str) -> Optional[str]:
        self.drop_until(match_pat)
        return self.pop()

    def drop_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> Optional[str]:
        while self.current_line_exists() and self.is_empty(empty_pat):
            self.pop()
        return self.get()

    def drop_until_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> Optional[str]:
        while self.current_line_exists() and not self.is_empty(empty_pat):
            self.pop()
        return self.get()

    def drop_until_empty_inclusive(
        self, empty_pat: str = EMPTY_STR_PATTERN
    ) -> Optional[str]:
        self.drop_until_empty(empty_pat)
        return self.pop()

    def pop_until(self, match_pat: str) -> Iterator[str]:
        while self.current_line_exists() and not self.match(match_pat):
            line = self.pop()
            if line is not None:
                yield line

    def pop_until_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> Iterator[str]:
        while self.current_line_exists() and not self.is_empty(empty_pat):
            line = self.pop()
            if line is not None:
                yield line


class ListReader(LinesReader):
    def __init__(self, lines: list[str]):
        self.contents = "\n".join(lines)
        self.raw_contents = self.contents
        self.lines: list[str] = lines
        self.n_lines = len(self.lines)
        self.current_line = 0


class CsvReader(LinesReader):
    def pop_csv_block_as_lines(self, empty_pat: str = EMPTY_STR_PATTERN) -> list[str]:
        self.drop_empty(empty_pat)
        lines = list(self.pop_until_empty(empty_pat))
        self.drop_empty(empty_pat)
        return lines

    def pop_csv_block_as_df(
        self,
        empty_pat: str = EMPTY_STR_PATTERN,
        *,
        header: Optional[int] = None,
        as_str: bool = False,
    ) -> Optional[pd.DataFrame]:
        if lines := self.pop_csv_block_as_lines(empty_pat):
            return pd.read_csv(
                StringIO("\n".join(lines)),
                header=header,
                dtype=str if as_str else None,
            )
        return None

    def drop_sections(self, match_pat: str) -> None:
        self.drop_empty()
        while self.match(match_pat):
            self.drop_until_empty()
            self.drop_empty()



import pytest

from allotropy.exceptions import AllotropeConversionError
from allotropy.parser_factory import Vendor
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_parser import (
    PerkinElmerEnvisionParser,
)
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from tests.parsers.perkin_elmer_envision.perkin_elmer_envision_data import (
    get_data,
    get_model,
)
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

OUTPUT_FILES = (
    "PE_Envision_absorbance_example01",
    "PE_Envision_fluorescence_example01",
    "PE_Envision_fluorescence_example02",
    "PE_Envision_fluorescence_example03",
    "PE_Envision_fluorescence_example04",
    "PE_Envision_luminescence_example01",
)

VENDOR_TYPE = Vendor.PERKIN_ELMER_ENVISION


@pytest.mark.parametrize("output_file", OUTPUT_FILES)
def test_parse_perkin_elmer_envision_to_asm(output_file: str) -> None:
    test_filepath = f"tests/parsers/perkin_elmer_envision/testdata/{output_file}.csv"
    expected_filepath = (
        f"tests/parsers/perkin_elmer_envision/testdata/{output_file}.json"
    )
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, "plate-reader/BENCHLING/2023/09/plate-reader.json")
    validate_contents(allotrope_dict, expected_filepath)


@pytest.mark.short
def test_get_model() -> None:
    parser = PerkinElmerEnvisionParser(TimestampParser())
    model = parser._get_model(get_data(), "file.txt")

    # remove all random UUIDs
    if agg_doc := model.plate_reader_aggregate_document:
        for i in range(len(plate_doc := agg_doc.plate_reader_document)):
            for j in range(
                len(
                    measurement_doc := plate_doc[
                        i
                    ].measurement_aggregate_document.measurement_document
                )
            ):
                measurement_doc[j].measurement_identifier = ""

    assert model == get_model()


def test_parse_missing_file() -> None:
    test_filepath = "tests/parsers/perkin_elmer_envision/testdata/PE_Envision_fluorescence_example01.tsv"
    with pytest.raises(
        AllotropeConversionError, match=f"File not found: {test_filepath}"
    ):
        from_file(test_filepath, VENDOR_TYPE)


def test_parse_incorrect_vendor() -> None:
    test_filepath = "tests/parsers/perkin_elmer_envision/testdata/PE_Envision_fluorescence_example01.csv"
    with pytest.raises(AllotropeConversionError):
        from_file(test_filepath, Vendor.AGILENT_GEN5)


def test_parse_file_missing_headers() -> None:
    test_filepath = (
        "tests/parsers/perkin_elmer_envision/testdata/example01_missing_header.csv"
    )
    with pytest.raises(AllotropeConversionError):
        from_file(test_filepath, VENDOR_TYPE)



import pytest

from allotropy.parser_factory import Vendor
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

OUTPUT_FILES = (
    "Demo_A260_dsDNA_Data",
    "Demo_A280_Protein",
)

VENDOR_TYPE = Vendor.UNCHAINED_LABS_LUNATIC
SCHEMA_FILE = "plate-reader/BENCHLING/2023/09/plate-reader.json"


@pytest.mark.parametrize("output_file", OUTPUT_FILES)
def test_parse_cedex_bioht_to_asm(output_file: str) -> None:
    test_filepath = f"tests/parsers/unchained_labs_lunatic/testdata/{output_file}.csv"
    expected_filepath = (
        f"tests/parsers/unchained_labs_lunatic/testdata/{output_file}.json"
    )
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, SCHEMA_FILE)
    validate_contents(allotrope_dict, expected_filepath)



from collections.abc import Collection
from typing import Any, Optional


class AllotropeConversionError(Exception):
    pass


def msg_for_error_on_unrecognized_value(
    key: str, value: Any, valid_values: Optional[Collection[Any]] = None
) -> str:
    msg = f"Unrecognized {key}: '{value}'."
    if valid_values:
        msg += f" Only {sorted(valid_values)} are supported."
    return msg



from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
import re
from typing import Any, Optional
import uuid

from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    DeviceControlAggregateDocument as DeviceControlAggregateDocumentFluorescence,
    DeviceControlDocumentItem as DeviceControlDocumentItemFluorescence,
    MeasurementAggregateDocument as MeasurementAggregateDocumentFluorescence,
    MeasurementDocumentItem as MeasurementDocumentItemFluorescence,
    Model as ModelFluorescence,
)
from allotropy.allotrope.models.luminescence_benchling_2023_09_luminescence import (
    DeviceControlAggregateDocument as DeviceControlAggregateDocumentLuminescence,
    DeviceControlDocumentItem as DeviceControlDocumentItemLuminescence,
    MeasurementAggregateDocument as MeasurementAggregateDocumentLuminescence,
    MeasurementDocumentItem as MeasurementDocumentItemLuminescence,
    Model as ModelLuminescence,
)
from allotropy.allotrope.models.shared.components.plate_reader import (
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueNanometer,
    TQuantityValueNumber,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TDatacube,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
)
from allotropy.allotrope.models.ultraviolet_absorbance_benchling_2023_09_ultraviolet_absorbance import (
    DeviceControlAggregateDocument as DeviceControlAggregateDocumentAbsorbance,
    DeviceControlDocumentItem as DeviceControlDocumentItemAbsorbance,
    MeasurementAggregateDocument as MeasurementAggregateDocumentAbsorbance,
    MeasurementDocumentItem as MeasurementDocumentItemAbsorbance,
    Model as ModelAbsorbance,
)
from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.values import (
    assert_not_none,
    natural_sort_key,
    PrimitiveValue,
    try_int,
    try_int_or_none,
    value_or_none,
)

BLOCKS_LINE_REGEX = r"^##BLOCKS=\s*(\d+)$"
END_LINE_REGEX = "~End"
EXPORT_VERSION = "1.3"
START_LETTER_CODE = ord("A")


def split_wavelengths(
    values: Optional[str], delimiter: str = " "
) -> Optional[list[int]]:
    if values is None:
        return None
    return [int(value) for value in values.split(delimiter)]


def get_well_coordinates(row_number: int, column_number: str) -> str:
    row_letters = ""
    while row_number > 0:
        row_number, remainder = divmod(row_number - 1, 26)
        row_letters = chr(START_LETTER_CODE + remainder) + row_letters
    return row_letters + str(column_number)


class ReadType(Enum):
    ENDPOINT = "Endpoint"
    KINETIC = "Kinetic"
    SPECTRUM = "Spectrum"
    WELL_SCAN = "Well Scan"


class ExportFormat(Enum):
    TIME_FORMAT = "TimeFormat"
    PLATE_FORMAT = "PlateFormat"


class DataType(Enum):
    RAW = "Raw"
    REDUCED = "Reduced"


@dataclass(frozen=True)
class Block:
    block_type: str
    raw_lines: list[str]

    @staticmethod
    def create(lines: list[str]) -> Block:
        block_cls_by_type: dict[str, type[Block]] = {
            "Group": GroupBlock,
            "Note": NoteBlock,
            "Plate": PlateBlock,
        }

        for key, cls in block_cls_by_type.items():
            if lines[0].startswith(key):
                return cls.create(lines)

        error = f"Expected block '{lines[0]}' to start with one of {sorted(block_cls_by_type.keys())}."
        raise AllotropeConversionError(error)


@dataclass(frozen=True)
class GroupBlock(Block):
    block_type: str
    name: str
    group_data: list[str]

    @staticmethod
    def create(raw_lines: list[str]) -> GroupBlock:
        group_data = []
        for i, line in enumerate(raw_lines):
            if line.startswith("Group Column"):
                group_data = raw_lines[1 : i - 1]
        # TODO handle group data

        return GroupBlock(
            block_type="Group",
            raw_lines=raw_lines,
            name=raw_lines[0][len("Group: ") :],
            group_data=group_data,
        )


# TODO do we need to do anything with these?
@dataclass(frozen=True)
class NoteBlock(Block):
    @staticmethod
    def create(raw_lines: list[str]) -> NoteBlock:
        return NoteBlock(block_type="Note", raw_lines=raw_lines)


@dataclass
class WellData:
    values: list[Optional[float]]
    dimensions: list[Optional[PrimitiveValue]]
    wavelengths: list[Optional[int]]
    temperature: Optional[str]
    processed_data: list[float]

    @staticmethod
    def create() -> WellData:
        return WellData(
            values=[],
            dimensions=[],
            wavelengths=[],
            temperature=None,
            processed_data=[],
        )

    def add_value(
        self,
        value: float,
        dimension: Optional[PrimitiveValue],
        temperature: Optional[str],
        wavelength: Optional[int],
    ) -> None:
        self.values.append(value)
        self.dimensions.append(dimension)
        self.wavelengths.append(wavelength)
        if temperature is not None:
            if self.temperature is not None and temperature != self.temperature:
                error = f"Expected all measurements to have the same temperature, but two have differing values of {self.temperature} and {temperature}."
                raise AllotropeConversionError(error)
            self.temperature = temperature

    @property
    def is_empty(self) -> bool:
        return not self.dimensions or not self.values


@dataclass(frozen=True)
class PlateBlock(Block):
    block_type: str
    name: Optional[str]
    export_format: Optional[str]
    read_type: Optional[str]
    data_type: Optional[str]
    kinetic_points: int
    num_wavelengths: Optional[int]
    wavelengths: list[int]
    num_columns: int
    num_wells: int
    well_data: defaultdict[str, WellData]
    data_header: list[Optional[str]]
    concept: str
    read_mode: str
    unit: str
    pmt_gain: Optional[str]
    num_rows: int
    excitation_wavelengths: Optional[list[int]]
    cutoff_filters: Optional[list[int]]

    @staticmethod
    def create(raw_lines: list[str]) -> PlateBlock:
        split_lines = [
            [value_or_none(value) for value in raw_line.split("\t")]
            for raw_line in raw_lines
        ]
        header = split_lines[0]
        read_mode = header[5]

        plate_block_cls: dict[str, type[PlateBlock]] = {
            "Absorbance": AbsorbancePlateBlock,
            "Fluorescence": FluorescencePlateBlock,
            "Luminescence": LuminescencePlateBlock,
        }

        cls = plate_block_cls.get(read_mode or "")
        if not cls:
            msg = msg_for_error_on_unrecognized_value(
                "read mode", read_mode, plate_block_cls.keys()
            )
            raise AllotropeConversionError(msg)

        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            _,  # Read mode
        ] = header[:6]
        if export_version != EXPORT_VERSION:
            error = f"Unsupported export version {export_version}; only {EXPORT_VERSION} is supported."
            raise AllotropeConversionError(error)

        data_type_idx = cls.get_data_type_idx()

        [
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            read_time_or_scan_pattern,
            read_interval_or_scan_density,
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            num_columns_raw,
            num_wells_raw,
        ] = header[data_type_idx : data_type_idx + 13]
        kinetic_points = try_int(kinetic_points_raw, "kinetic_points")
        num_columns = try_int(num_columns_raw, "num_columns")
        num_wells = try_int(num_wells_raw, "num_wells")
        num_wavelengths = try_int_or_none(num_wavelengths_raw)
        wavelengths = split_wavelengths(wavelengths_str) or []

        extra_attr = cls.parse_read_mode_header(header)

        well_data: defaultdict[str, WellData] = defaultdict(WellData.create)
        data_header = split_lines[1]
        data_lines = split_lines[2:]
        if export_format == ExportFormat.TIME_FORMAT.value:
            PlateBlock._parse_time_format_data(
                wavelengths,
                read_type,
                well_data,
                kinetic_points,
                num_wells,
                data_header,
                data_type,
                num_wavelengths,
                data_lines,
            )
        elif export_format == ExportFormat.PLATE_FORMAT.value:
            PlateBlock._parse_plate_format_data(
                wavelengths,
                read_type,
                well_data,
                data_type,
                kinetic_points,
                extra_attr.num_rows,
                num_wavelengths,
                num_columns,
                data_header,
                data_lines,
            )
        else:
            msg = msg_for_error_on_unrecognized_value(
                "export format", export_format, ExportFormat._member_names_
            )
            raise AllotropeConversionError(msg)

        return cls(
            block_type="Plate",
            raw_lines=raw_lines,
            name=name,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=kinetic_points,
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            well_data=well_data,
            data_header=data_header,
            concept=extra_attr.concept,
            read_mode=extra_attr.read_mode,
            unit=extra_attr.unit,
            pmt_gain=extra_attr.pmt_gain,
            num_rows=extra_attr.num_rows,
            excitation_wavelengths=extra_attr.excitation_wavelengths,
            cutoff_filters=extra_attr.cutoff_filters,
        )

    @staticmethod
    def _parse_reduced_plate_rows(
        num_columns: int,
        data_header: list[Optional[str]],
        well_data: defaultdict[str, WellData],
        reduced_data_rows: list[list[Optional[str]]],
    ) -> None:
        for i, row in enumerate(reduced_data_rows):
            for j, value in enumerate(row[2 : num_columns + 2]):
                if value is None:
                    continue
                col_number = assert_not_none(data_header[j + 2], "column number")
                well = get_well_coordinates(i + 1, col_number)
                well_data[well].processed_data.append(float(value))

    @staticmethod
    def _parse_reduced_columns(
        data_header: list[Optional[str]],
        well_data: defaultdict[str, WellData],
        reduced_data_row: list[Optional[str]],
    ) -> None:
        for i, value in enumerate(reduced_data_row[2:]):
            if value is None:
                continue
            well = assert_not_none(data_header[i + 2], "well")
            well_data[well].processed_data.append(float(value))

    @staticmethod
    def _add_data_point(
        read_type: Optional[str],
        well_data: defaultdict[str, WellData],
        well: str,
        value: str,
        data_key: Optional[str],
        temperature: Optional[str],
        wavelength: Optional[int],
    ) -> None:
        dimension = wavelength if read_type == ReadType.ENDPOINT.value else data_key
        well_data[well].add_value(
            value=float(value),
            dimension=dimension,
            temperature=temperature,
            wavelength=wavelength,
        )

    @staticmethod
    def _parse_time_format_data(
        wavelengths: list[int],
        read_type: Optional[str],
        well_data: defaultdict[str, WellData],
        kinetic_points: int,
        num_wells: int,
        data_header: list[Optional[str]],
        data_type: Optional[str],
        num_wavelengths: Optional[int],
        data_lines: list[list[Optional[str]]],
    ) -> None:
        if data_type == DataType.RAW.value:
            num_row_blocks = num_wavelengths if num_wavelengths is not None else 1
            for wavelength_index in range(num_row_blocks):
                start_index = wavelength_index * (kinetic_points + 1)
                wavelength_rows = data_lines[start_index : start_index + kinetic_points]
                for row in wavelength_rows:
                    for i, value in enumerate(row[2 : num_wells + 2]):
                        if value is None:
                            continue
                        well = assert_not_none(data_header[i + 2], "well")
                        PlateBlock._add_data_point(
                            read_type,
                            well_data,
                            well,
                            value,
                            data_key=row[0],
                            temperature=row[1],
                            wavelength=PlateBlock.get_wavelength(
                                read_type, wavelengths, wavelength_index
                            ),
                        )
            if len(data_lines) > (kinetic_points + 1) * num_row_blocks:
                reduced_row = data_lines[-1][: num_wells + 2]
                PlateBlock._parse_reduced_columns(data_header, well_data, reduced_row)
        elif data_type == DataType.REDUCED.value:
            reduced_row = data_lines[-1][: num_wells + 2]
            PlateBlock._parse_reduced_columns(data_header, well_data, reduced_row)
        else:
            msg = msg_for_error_on_unrecognized_value(
                "data type", data_type, DataType._member_names_
            )
            raise AllotropeConversionError(msg)

    @staticmethod
    def _parse_plate_format_data(
        wavelengths: list[int],
        read_type: Optional[str],
        well_data: defaultdict[str, WellData],
        data_type: Optional[str],
        kinetic_points: int,
        num_rows: int,
        num_wavelengths: Optional[int],
        num_columns: int,
        data_header: list[Optional[str]],
        data_lines: list[list[Optional[str]]],
    ) -> None:
        end_raw_data_index = 0
        if data_type == DataType.RAW.value:
            for read_index in range(kinetic_points):
                start_index = read_index * (num_rows + 1)
                read_rows = data_lines[start_index : start_index + num_rows]
                data_key = read_rows[0][0]
                temperature = read_rows[0][1]
                for i, row in enumerate(read_rows):
                    wavelength_index = 0
                    num_row_blocks = (
                        num_wavelengths if num_wavelengths is not None else 1
                    )
                    for wavelength_index in range(num_row_blocks):
                        col_start_index = 2 + (wavelength_index * (num_columns + 1))
                        for j, value in enumerate(
                            row[col_start_index : col_start_index + num_columns]
                        ):
                            if value is None:
                                continue
                            col_number = assert_not_none(
                                data_header[j + 2], "column number"
                            )
                            well = get_well_coordinates(i + 1, col_number)
                            PlateBlock._add_data_point(
                                read_type,
                                well_data,
                                well,
                                value,
                                data_key=data_key,
                                temperature=temperature,
                                wavelength=PlateBlock.get_wavelength(
                                    read_type, wavelengths, wavelength_index
                                ),
                            )
            end_raw_data_index = ((num_rows + 1) * kinetic_points) + 1
            reduced_data_rows = data_lines[end_raw_data_index:]
            if len(reduced_data_rows) == num_rows:
                PlateBlock._parse_reduced_plate_rows(
                    num_columns,
                    data_header,
                    well_data,
                    reduced_data_rows,
                )
        elif data_type == DataType.REDUCED.value:
            reduced_data_rows = data_lines[end_raw_data_index:]
            PlateBlock._parse_reduced_plate_rows(
                num_columns, data_header, well_data, reduced_data_rows
            )
        else:
            msg = msg_for_error_on_unrecognized_value(
                "data type", data_type, DataType._member_names_
            )
            raise AllotropeConversionError(msg)

    @staticmethod
    def parse_read_mode_header(header: list[Optional[str]]) -> PlateBlockExtraAttr:
        raise NotImplementedError

    @staticmethod
    def get_data_type_idx() -> int:
        raise NotImplementedError

    @property
    def is_single_wavelength(self) -> bool:
        return (
            self.read_type in {ReadType.SPECTRUM.value, ReadType.ENDPOINT.value}
            and len(self.wavelengths) == 1
        )

    @property
    def has_wavelength_dimension(self) -> bool:
        return (
            self.read_type not in {ReadType.SPECTRUM.value, ReadType.ENDPOINT.value}
            and len(self.wavelengths) > 1
        )

    @staticmethod
    def get_wavelength(
        read_type: Optional[str],
        wavelengths: list[int],
        wavelength_index: int,
    ) -> Optional[int]:
        if read_type == ReadType.SPECTRUM.value:
            return None
        else:
            return wavelengths[wavelength_index] if wavelengths else None

    def get_data_cube_dimensions(self) -> list[tuple[str, str, Optional[str]]]:
        dimensions: list[tuple[str, str, Optional[str]]] = []
        if self.read_type == ReadType.KINETIC.value:
            dimensions = [("double", "elapsed time", "s")]
        elif self.read_type == ReadType.WELL_SCAN.value:
            dimensions = [("int", "x", None)]
        elif self.read_type in {ReadType.SPECTRUM.value, ReadType.ENDPOINT.value}:
            dimensions = [("int", "wavelength", "nm")]
        else:
            error = f"Cannot make data cube for read type {self.read_type}; only {sorted(ReadType._member_names_)} are supported."
            raise AllotropeConversionError(error)
        if self.has_wavelength_dimension:
            dimensions.append(("int", "wavelength", "nm"))
        return dimensions

    def generate_sample_document(self, well: str) -> SampleDocument:
        return SampleDocument(well_location_identifier=well, plate_barcode=self.name)

    def generate_data_cube(self, well_data: WellData) -> TDatacube:
        dimension_data = [well_data.dimensions] + (
            [well_data.wavelengths] if self.has_wavelength_dimension else []
        )
        return TDatacube(
            cube_structure=TDatacubeStructure(
                [
                    TDatacubeComponent(FieldComponentDatatype(data_type), concept, unit)
                    for data_type, concept, unit in self.get_data_cube_dimensions()
                ],
                [
                    TDatacubeComponent(
                        FieldComponentDatatype("double"), self.concept, self.unit
                    )
                ],
            ),
            data=TDatacubeData(
                dimension_data,  # type: ignore[arg-type]
                [well_data.values],
            ),
        )

    def generate_processed_data_aggreate_document(
        self, well_data: WellData
    ) -> ProcessedDataAggregateDocument:
        return ProcessedDataAggregateDocument(
            [
                ProcessedDataDocumentItem(
                    val, data_processing_description="processed data"
                )
                for val in well_data.processed_data
            ]
        )

    def to_allotrope(self) -> Any:
        raise NotImplementedError


@dataclass(frozen=True)
class PlateBlockExtraAttr:
    concept: str
    read_mode: str
    unit: str
    pmt_gain: Optional[str]
    num_rows: int
    excitation_wavelengths: Optional[list[int]]
    cutoff_filters: Optional[list[int]]


@dataclass(frozen=True)
class FluorescencePlateBlock(PlateBlock):
    EXCITATION_WAVELENGTHS_IDX: int = 20

    @staticmethod
    def get_data_type_idx() -> int:
        return 7

    @staticmethod
    def parse_read_mode_header(header: list[Optional[str]]) -> PlateBlockExtraAttr:
        [
            excitation_wavelengths_str,
            _,  # cutoff
            cutoff_filters_str,
            _,  # sweep_wave
            _,  # sweep_wavelength
            _,  # reads_per_well
            pmt_gain,
            _,  # start_integration_time
            _,  # end_integration_time
            _,  # first_row
            num_rows,
        ] = header[
            FluorescencePlateBlock.EXCITATION_WAVELENGTHS_IDX : FluorescencePlateBlock.EXCITATION_WAVELENGTHS_IDX
            + 11
        ]

        return PlateBlockExtraAttr(
            concept="fluorescence",
            read_mode="Fluorescence",
            unit="RFU",
            pmt_gain=pmt_gain,
            num_rows=try_int(num_rows, "num_rows"),
            excitation_wavelengths=split_wavelengths(excitation_wavelengths_str),
            cutoff_filters=split_wavelengths(cutoff_filters_str),
        )

    # TODO: the reason we can't factor out DeviceControlDocumentItemFluorescence and the enclosing classes is because the
    # Fluorescence ASM model has these extra fields. We may be able to fix this by templating the PlateBlock
    # class on these classes, or some clever shared/models refactoring.
    def generate_device_control_doc(self) -> DeviceControlDocumentItemFluorescence:
        device_control_doc = DeviceControlDocumentItemFluorescence(
            detector_gain_setting=self.pmt_gain
        )

        if self.is_single_wavelength:
            device_control_doc.detector_wavelength_setting = TQuantityValueNanometer(
                self.wavelengths[0]
            )
        if self.excitation_wavelengths and len(self.excitation_wavelengths) == 1:
            device_control_doc.excitation_wavelength_setting = TQuantityValueNanometer(
                self.excitation_wavelengths[0]
            )
        if self.cutoff_filters and len(self.cutoff_filters) == 1:
            device_control_doc.wavelength_filter_cutoff_setting = (
                TQuantityValueNanometer(self.cutoff_filters[0])
            )

        return device_control_doc

    def generate_measurement_doc(
        self, well: str, well_data: WellData
    ) -> MeasurementDocumentItemFluorescence:
        measurement = MeasurementDocumentItemFluorescence(
            DeviceControlAggregateDocumentFluorescence(
                [self.generate_device_control_doc()]
            ),
            self.generate_sample_document(well),
        )

        if well_data.temperature is not None:
            measurement.compartment_temperature = TQuantityValueDegreeCelsius(
                float(well_data.temperature)
            )

        if not well_data.is_empty:
            measurement.data_cube = self.generate_data_cube(well_data)

        if well_data.processed_data:
            measurement.processed_data_aggregate_document = (
                self.generate_processed_data_aggreate_document(well_data)
            )

        return measurement

    def to_allotrope(self) -> Any:
        wells = sorted(self.well_data.keys(), key=natural_sort_key)

        allotrope_file = ModelFluorescence(
            measurement_aggregate_document=MeasurementAggregateDocumentFluorescence(
                measurement_identifier=str(uuid.uuid4()),
                plate_well_count=TQuantityValueNumber(self.num_wells),
                measurement_document=[
                    self.generate_measurement_doc(well, self.well_data[well])
                    for well in wells
                ],
            )
        )

        return allotrope_file


@dataclass(frozen=True)
class LuminescencePlateBlock(PlateBlock):
    EXCITATION_WAVELENGTHS_IDX: int = 19

    @staticmethod
    def get_data_type_idx() -> int:
        return 6

    @staticmethod
    def parse_read_mode_header(header: list[Optional[str]]) -> PlateBlockExtraAttr:
        [
            excitation_wavelengths_str,
            _,  # cutoff
            cutoff_filters_str,
            _,  # sweep_wave
            _,  # sweep_wavelength
            _,  # reads_per_well
            pmt_gain,
            _,  # start_integration_time
            _,  # end_integration_time
            _,  # first_row
            num_rows,
        ] = header[
            LuminescencePlateBlock.EXCITATION_WAVELENGTHS_IDX : LuminescencePlateBlock.EXCITATION_WAVELENGTHS_IDX
            + 11
        ]
        return PlateBlockExtraAttr(
            concept="luminescence",
            read_mode="Luminescence",
            unit="RLU",
            pmt_gain=pmt_gain,
            num_rows=try_int(num_rows, "num_rows"),
            excitation_wavelengths=split_wavelengths(excitation_wavelengths_str),
            cutoff_filters=split_wavelengths(cutoff_filters_str),
        )

    def generate_device_control_doc(self) -> DeviceControlDocumentItemLuminescence:
        device_control_doc = DeviceControlDocumentItemLuminescence(
            detector_gain_setting=self.pmt_gain
        )

        if self.is_single_wavelength:
            device_control_doc.detector_wavelength_setting = TQuantityValueNanometer(
                self.wavelengths[0]
            )

        return device_control_doc

    def generate_measurement_doc(
        self, well: str, well_data: WellData
    ) -> MeasurementDocumentItemLuminescence:
        measurement = MeasurementDocumentItemLuminescence(
            DeviceControlAggregateDocumentLuminescence(
                [self.generate_device_control_doc()]
            ),
            self.generate_sample_document(well),
        )

        if well_data.temperature is not None:
            measurement.compartment_temperature = TQuantityValueDegreeCelsius(
                float(well_data.temperature)
            )

        if not well_data.is_empty:
            measurement.data_cube = self.generate_data_cube(well_data)

        if well_data.processed_data:
            measurement.processed_data_aggregate_document = (
                self.generate_processed_data_aggreate_document(well_data)
            )

        return measurement

    def to_allotrope(self) -> Any:
        wells = sorted(self.well_data.keys(), key=natural_sort_key)

        allotrope_file = ModelLuminescence(
            measurement_aggregate_document=MeasurementAggregateDocumentLuminescence(
                measurement_identifier=str(uuid.uuid4()),
                plate_well_count=TQuantityValueNumber(self.num_wells),
                measurement_document=[
                    self.generate_measurement_doc(well, self.well_data[well])
                    for well in wells
                ],
            )
        )

        return allotrope_file


@dataclass(frozen=True)
class AbsorbancePlateBlock(PlateBlock):
    @staticmethod
    def get_data_type_idx() -> int:
        return 6

    @staticmethod
    def parse_read_mode_header(header: list[Optional[str]]) -> PlateBlockExtraAttr:
        return PlateBlockExtraAttr(
            concept="absorbance",
            read_mode="Absorbance",
            unit="mAU",
            pmt_gain=None,
            num_rows=try_int(header[20], "num_rows"),
            excitation_wavelengths=None,
            cutoff_filters=None,
        )

    def generate_device_control_doc(self) -> DeviceControlDocumentItemAbsorbance:
        device_control_doc = DeviceControlDocumentItemAbsorbance(
            detector_gain_setting=self.pmt_gain
        )

        if self.is_single_wavelength:
            device_control_doc.detector_wavelength_setting = TQuantityValueNanometer(
                self.wavelengths[0]
            )

        return device_control_doc

    def generate_measurement_doc(
        self, well: str, well_data: WellData
    ) -> MeasurementDocumentItemAbsorbance:
        measurement = MeasurementDocumentItemAbsorbance(
            DeviceControlAggregateDocumentAbsorbance(
                [self.generate_device_control_doc()]
            ),
            self.generate_sample_document(well),
        )

        if well_data.temperature is not None:
            measurement.compartment_temperature = TQuantityValueDegreeCelsius(
                float(well_data.temperature)
            )

        if not well_data.is_empty:
            measurement.data_cube = self.generate_data_cube(well_data)

        if well_data.processed_data:
            measurement.processed_data_aggregate_document = (
                self.generate_processed_data_aggreate_document(well_data)
            )

        return measurement

    def to_allotrope(self) -> Any:
        wells = sorted(self.well_data.keys(), key=natural_sort_key)

        allotrope_file = ModelAbsorbance(
            measurement_aggregate_document=MeasurementAggregateDocumentAbsorbance(
                measurement_identifier=str(uuid.uuid4()),
                plate_well_count=TQuantityValueNumber(self.num_wells),
                measurement_document=[
                    self.generate_measurement_doc(well, self.well_data[well])
                    for well in wells
                ],
            )
        )

        return allotrope_file


@dataclass(frozen=True)
class BlockList:
    blocks: list[Block]

    @staticmethod
    def create(lines_reader: LinesReader) -> BlockList:
        return BlockList(
            blocks=[
                Block.create(block) for block in BlockList._iter_blocks(lines_reader)
            ]
        )

    @staticmethod
    def _get_n_blocks(lines_reader: LinesReader) -> int:
        start_line = lines_reader.pop() or ""
        if search_result := re.search(BLOCKS_LINE_REGEX, start_line):
            return int(search_result.group(1))
        msg = msg_for_error_on_unrecognized_value("start line", start_line)
        raise AllotropeConversionError(msg)

    @staticmethod
    def _iter_blocks(lines_reader: LinesReader) -> Iterator[list[str]]:
        n_blocks = BlockList._get_n_blocks(lines_reader)
        for _ in range(n_blocks):
            yield list(lines_reader.pop_until(END_LINE_REGEX))
            lines_reader.pop()  # drop end line
            lines_reader.drop_empty()


@dataclass(frozen=True)
class Data:
    block_list: BlockList

    def get_plate_block(self) -> PlateBlock:
        plate_blocks = [
            block for block in self.block_list.blocks if isinstance(block, PlateBlock)
        ]

        if len(plate_blocks) != 1:
            block_types = [block.block_type for block in self.block_list.blocks]
            block_counts = {bt: block_types.count(bt) for bt in set(block_types)}
            error = f"Expected exactly 1 plate block; got {block_counts}."
            raise AllotropeConversionError(error)

        return plate_blocks[0]

    @staticmethod
    def create(lines_reader: LinesReader) -> Data:
        return Data(block_list=BlockList.create(lines_reader))



from typing import Optional

import pytest

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import (
    assert_not_none,
    natural_sort_key,
    try_float,
    try_float_or_none,
    try_int,
    try_int_or_none,
    value_or_none,
)


def test_assert_not_none() -> None:
    x = 3
    assert assert_not_none(x) == x


def test_assert_not_none_fails() -> None:
    with pytest.raises(Exception, match="^Expected non-null value.$"):
        assert_not_none(None)


def test_assert_not_none_fails_with_message() -> None:
    with pytest.raises(Exception, match="^param_name was None$"):
        assert_not_none(None, msg="param_name was None")


def test_assert_not_none_fails_with_name() -> None:
    with pytest.raises(Exception, match="^Expected non-null value for param_name.$"):
        assert_not_none(None, "param_name")


def test_assert_not_none_fails_with_message_and_name() -> None:
    with pytest.raises(Exception, match="^param_name was None$"):
        assert_not_none(None, "param_name", "param_name was None")


@pytest.mark.short
@pytest.mark.parametrize(
    "key,expected",
    [
        ("", []),
        ("a", ["a"]),
        ("1", ["         1"]),
        ("a1b2c", ["a", "         1", "b", "         2", "c"]),
    ],
)
def test_natural_sort_key(key: str, expected: list[str]) -> None:
    assert natural_sort_key(key) == expected


def _try_float(value: Optional[str]) -> float:
    return try_float(value, "param")


def test_try_float() -> None:
    assert _try_float("1.0") == 1.0


@pytest.mark.short
@pytest.mark.parametrize(
    "value,expected_regex",
    [
        (None, "Expected non-null value for param."),
        ("a", "Invalid float string: 'a'."),
    ],
)
def test_try_float_fails(value: Optional[str], expected_regex: str) -> None:
    with pytest.raises(AllotropeConversionError, match=expected_regex):
        _try_float(value)


@pytest.mark.short
@pytest.mark.parametrize(
    "value,expected",
    [
        (None, None),
        ("", None),
        ("1", 1),
        ("1.1", 1.1),
    ],
)
def test_try_float_or_none(value: Optional[str], expected: Optional[int]) -> None:
    assert try_float_or_none(value) == expected


def _try_int(value: Optional[str]) -> int:
    return try_int(value, "param")


def test_try_int() -> None:
    assert _try_int("1") == 1


@pytest.mark.short
@pytest.mark.parametrize(
    "value,expected_regex",
    [
        (None, "Expected non-null value for param."),
        ("a", "Invalid integer string: 'a'."),
    ],
)
def test_try_int_fails(value: Optional[str], expected_regex: str) -> None:
    with pytest.raises(AllotropeConversionError, match=expected_regex):
        _try_int(value)


@pytest.mark.short
@pytest.mark.parametrize(
    "value,expected",
    [
        (None, None),
        ("", None),
        ("1", 1),
        ("1.1", None),
    ],
)
def test_try_int_or_none(value: Optional[str], expected: Optional[float]) -> None:
    assert try_int_or_none(value) == expected


@pytest.mark.short
@pytest.mark.parametrize(
    "value,expected",
    [
        ("", None),
        ("  ", None),
        (" 1 ", "1"),
    ],
)
def test_value_or_none(value: str, expected: Optional[str]) -> None:
    assert value_or_none(value) == expected



from more_itertools import one
import pytest

from allotropy.allotrope.models.cell_counting_benchling_2023_11_cell_counting import (
    Model,
)
from allotropy.parser_factory import Vendor
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_parser import ViCellBluParser
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from tests.parsers.beckman_vi_cell_blu.vi_cell_blu_data import (
    get_data,
    get_filename,
    get_model,
)
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

OUTPUT_FILES = (
    "Beckman_Vi-Cell-BLU_example01",
    "Beckman_Vi-Cell-BLU_example02",
)

VENDOR_TYPE = Vendor.BECKMAN_VI_CELL_BLU
SCHEMA_FILE = "cell-counting/BENCHLING/2023/11/cell-counting.json"
TEST_DATA_DIR = "tests/parsers/beckman_vi_cell_blu/testdata/"


def _get_test_file_path(output_file: str) -> str:
    return f"{TEST_DATA_DIR}/{output_file}.csv"


def _get_expected_file_path(output_file: str) -> str:
    return f"{TEST_DATA_DIR}/{output_file}.json"


@pytest.mark.parametrize("output_file", OUTPUT_FILES)
def test_parse_vi_cell_blu_to_asm_schema_is_valid(output_file: str) -> None:
    test_filepath = _get_test_file_path(output_file)
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, SCHEMA_FILE)


@pytest.mark.parametrize("output_file", OUTPUT_FILES)
def test_parse_vi_cell_blu_to_asm_expected_contents(output_file: str) -> None:
    test_filepath = _get_test_file_path(output_file)
    expected_filepath = _get_expected_file_path(output_file)
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_contents(allotrope_dict, expected_filepath)


def _clear_measurement_identifier(model: Model) -> None:
    cell_counting_aggregate_document = model.cell_counting_aggregate_document
    assert cell_counting_aggregate_document

    cell_counting_document_item = one(
        cell_counting_aggregate_document.cell_counting_document
    )
    measurement_document = one(
        cell_counting_document_item.measurement_aggregate_document.measurement_document
    )
    measurement_document.measurement_identifier = ""


def test_get_model() -> None:
    parser = ViCellBluParser(TimestampParser())
    result = parser._get_model(get_data(), get_filename())
    _clear_measurement_identifier(result)
    assert result == get_model()



from typing import Any

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.moldev_softmax_pro.softmax_pro_structure import Data
from allotropy.parsers.vendor_parser import VendorParser


class SoftmaxproParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Any:
        contents = named_file_contents.contents
        lines_reader = LinesReader(contents)
        data = Data.create(lines_reader)
        return self._get_model(data)

    def _get_model(self, data: Data) -> Any:
        return data.get_plate_block().to_allotrope()



from abc import ABC, abstractmethod
from typing import Any

from allotropy.allotrope.models.shared.definitions.definitions import TDateTimeValue
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from allotropy.parsers.utils.values import assert_not_none


class VendorParser(ABC):
    timestamp_parser: TimestampParser

    """Base class for all vendor parsers."""

    def __init__(self, timestamp_parser: TimestampParser):
        self.timestamp_parser = assert_not_none(timestamp_parser, "timestamp_parser")

    @abstractmethod
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Any:
        raise NotImplementedError

    # TODO: make time param a str
    def _get_date_time(self, time: Any) -> TDateTimeValue:
        assert_not_none(time, "time")

        return self.timestamp_parser.parse(str(time))



import pytest

from allotropy.parser_factory import Vendor
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

valid_files = (
    "Weyland_Yutani_simple_correct",
    "Weyland_Yutani_checksum_correct",
)

SCHEMA = "fluorescence/BENCHLING/2023/09/fluorescence.json"
TESTDATA = "tests/parsers/example_weyland_yutani/testdata"
VENDOR_TYPE = Vendor.EXAMPLE_WEYLAND_YUTANI


@pytest.mark.parametrize("filestem", valid_files)
def test_parse_weyland_yutani_to_asm(filestem: str) -> None:
    test_filepath = f"{TESTDATA}/{filestem}.csv"
    expected_filepath = f"{TESTDATA}/{filestem}.json"
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, SCHEMA)
    validate_contents(allotrope_dict, expected_filepath)



import re

import pandas as pd
import pytest

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.unchained_labs_lunatic.constants import (
    INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG,
    NO_DATE_OR_TIME_ERROR_MSG,
    NO_MEASUREMENT_IN_PLATE_ERROR_MSG,
)
from allotropy.parsers.unchained_labs_lunatic.unchained_labs_lunatic_structure import (
    Data,
    Measurement,
    WellPlate,
)


@pytest.mark.parametrize(
    "wavelength,absorbance_value,sample_identifier,location_identifier,well_plate_identifier",
    [
        (245, 590, "Sample Name 1", "plateID1", "A1"),
        (230, 90, "Sample Name 2", "plateID2", "A3"),
        (450, 34.9, "Sample Name 3", "plateID3", None),
    ],
)
@pytest.mark.short
def test_create_measurement(
    wavelength: int,
    absorbance_value: float,
    sample_identifier: str,
    location_identifier: str,
    well_plate_identifier: str,
) -> None:
    wavelength_column = f"A{wavelength}"
    well_plate_data = {
        wavelength_column: absorbance_value,
        "Sample name": sample_identifier,
        "Plate ID": location_identifier,
        "Plate Position": well_plate_identifier,
    }
    measurement = Measurement.create(pd.Series(well_plate_data), wavelength_column)

    assert measurement.wavelength == wavelength
    assert measurement.absorbance == absorbance_value
    assert measurement.sample_identifier == sample_identifier
    assert measurement.location_identifier == location_identifier
    assert measurement.well_plate_identifier == well_plate_identifier


@pytest.mark.short
def test_create_measurement_with_no_wavelength_column() -> None:
    well_plate_data = {
        "Sample name": "dummy name",
        "Plate ID": "dummy ID",
        "Plate Position": "B3",
    }
    wavelength_column = "A250"
    msg = NO_MEASUREMENT_IN_PLATE_ERROR_MSG.format(wavelength_column)
    with pytest.raises(AllotropeConversionError, match=msg):
        Measurement.create(pd.Series(well_plate_data), wavelength_column)


@pytest.mark.short
def test_create_title_with_incorrect_wavelength_column_format() -> None:
    msg = INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG
    well_plate_data = pd.Series({"Sample name": "dummy name"})
    with pytest.raises(AllotropeConversionError, match=re.escape(msg)):
        Measurement.create(well_plate_data, "Sample name")


@pytest.mark.short
def test_create_well_plate() -> None:
    analytical_method_identifier = "dummy method"
    date = "17/10/2016"
    time = "7:19:18"
    plate_data = {
        "A250": 23.45,
        "Sample name": "dummy name",
        "Plate ID": "dummy ID",
        "Application": analytical_method_identifier,
        "Date": date,
        "Time": time,
    }
    well_plate = WellPlate.create(pd.Series(plate_data), ["A250"])
    assert well_plate.analytical_method_identifier == analytical_method_identifier
    assert well_plate.measurement_time == f"{date} {time}"
    assert well_plate.measurements[0].absorbance == 23.45


@pytest.mark.short
def test_create_well_plate_with_two_measurements() -> None:
    plate_data = {
        "A452": 23.45,
        "A280": 23.45,
        "Sample name": "dummy name",
        "Plate ID": "dummy ID",
        "Date": "17/10/2016",
        "Time": "7:19:18",
    }
    well_plate = WellPlate.create(pd.Series(plate_data), ["A452", "A280"])

    assert len(well_plate.measurements) == 2


@pytest.mark.short
def test_create_well_plate_without_date_column_then_raise() -> None:
    plate_data = {
        "Sample name": "dummy name",
        "Plate ID": "dummy ID",
        "Time": "7:19:18",
    }
    with pytest.raises(AllotropeConversionError, match=NO_DATE_OR_TIME_ERROR_MSG):
        WellPlate.create(pd.Series(plate_data), [])


@pytest.mark.short
def test_create_data() -> None:
    plate_data = {
        "Sample name": ["batch_id", "batch_id", ""],
        "Plate ID": ["Plate1", "Plate1", "Plate1"],
        "Application": ["dummyApp", "dummyApp", "dummyApp"],
        "Date": ["2021-05-20", "2021-05-20", "2023-05-20"],
        "Time": ["16:55:51", "16:56:51", "16:55:51"],
        "Instrument ID": [14, 14, 14],
        "A250": [23.4, 32.6, 439],
    }
    data = Data.create(pd.DataFrame(plate_data))

    assert data.device_identifier == "14"
    assert len(data.well_plate_list) == 3



import re
from typing import Optional, TypeVar, Union

import pandas as pd

from allotropy.exceptions import AllotropeConversionError

PrimitiveValue = Union[str, int, float]


def try_int(value: Optional[str], value_name: str) -> int:
    try:
        return int(assert_not_none(value, value_name))
    except ValueError as e:
        msg = f"Invalid integer string: '{value}'."
        raise AllotropeConversionError(msg) from e


def try_int_or_none(value: Optional[str]) -> Optional[int]:
    try:
        return int(value or "")
    except ValueError:
        return None


def try_float(value: Optional[str], value_name: str) -> float:
    try:
        return float(assert_not_none(value, value_name))
    except ValueError as e:
        msg = f"Invalid float string: '{value}'."
        raise AllotropeConversionError(msg) from e


def try_float_or_none(value: Optional[str]) -> Optional[float]:
    try:
        return float(value or "")
    except ValueError:
        return None


def natural_sort_key(key: str) -> list[str]:
    """Returns a sort key that treats numeric substrings as parsed integers for comparison."""
    tokens = [token for token in re.split(r"(\d+)", key) if token]
    return [
        f"{int(token):>10}" if token.isdecimal() else token.lower() for token in tokens
    ]


T = TypeVar("T")


def assert_not_none(
    value: Optional[T], name: Optional[str] = None, msg: Optional[str] = None
) -> T:
    if value is None:
        error = msg or f"Expected non-null value{f' for {name}' if name else ''}."
        raise AllotropeConversionError(error)
    return value


def value_or_none(value: str) -> Optional[str]:
    return value.strip() or None


def df_to_series(
    df: pd.DataFrame,
    msg: str,
) -> pd.Series:  # type: ignore[type-arg]
    n_rows, _ = df.shape
    if n_rows == 1:
        return pd.Series(df.iloc[0], index=df.columns)
    raise AllotropeConversionError(msg)


def assert_not_empty_df(df: pd.DataFrame, msg: str) -> pd.DataFrame:
    if df.empty:
        raise AllotropeConversionError(msg)
    return df


def try_str_from_series_or_none(
    data: pd.Series,  # type: ignore[type-arg]
    key: str,
    default: Optional[str] = None,
) -> Optional[str]:
    value = data.get(key, default)
    return None if value is None else str(value)


def try_str_from_series(
    series: pd.Series,  # type: ignore[type-arg]
    key: str,
    msg: Optional[str] = None,
) -> str:
    return assert_not_none(try_str_from_series_or_none(series, key), key, msg)


def try_int_from_series_or_none(
    data: pd.Series,  # type: ignore[type-arg]
    key: str,
) -> Optional[int]:
    try:
        value = data.get(key)
        return try_int(str(value), key)
    except Exception as e:
        msg = f"Unable to convert '{value}' (with key '{key}') to integer value."
        raise AllotropeConversionError(msg) from e


def try_int_from_series(
    data: pd.Series,  # type: ignore[type-arg]
    key: str,
    msg: Optional[str] = None,
) -> int:
    return assert_not_none(try_int_from_series_or_none(data, key), key, msg)


def try_float_from_series_or_none(
    data: pd.Series,  # type: ignore[type-arg]
    key: str,
) -> Optional[float]:
    try:
        value = data.get(key)
        return try_float_or_none(str(value))
    except Exception as e:
        msg = f"Unable to convert '{value}' (with key '{key}') to float value."
        raise AllotropeConversionError(msg) from e


def try_float_from_series(
    data: pd.Series,  # type: ignore[type-arg]
    key: str,
    msg: Optional[str] = None,
) -> float:
    return assert_not_none(try_float_from_series_or_none(data, key), key, msg)


def try_bool_from_series_or_none(
    data: pd.Series,  # type: ignore[type-arg]
    key: str,
) -> Optional[bool]:
    try:
        value = data.get(key)
        return None if value is None else bool(value)
    except Exception as e:
        msg = f"Unable to convert '{value}' (with key '{key}') to boolean value."
        raise AllotropeConversionError(msg) from e
